from __future__ import annotations

import configparser
import logging
import os
from typing import TYPE_CHECKING, Iterable

if TYPE_CHECKING:
    from _typeshed import StrOrBytesPath

logger = logging.getLogger(__name__)


class ConfigParserOverride:
    def __init__(
        self,
        env_prefix: str = "",
        **overrides,
    ):
        """
        Initialize the ConfigParserOverride.

        :param env_prefix: Optional prefix for environment variables, defaults to an
            empty string.
        :type env_prefix: str, optional
        :param overrides: Keyword arguments to directly override configuration values.
        """
        self._config = configparser.ConfigParser()
        self.env_prefix = env_prefix
        self.overrides = overrides

    def _find_and_override(self):
        """
        Internal method to override configuration values with environment variables.

        This method iterates over all sections and keys in the configuration, and if a
        corresponding environment variable is set, it overrides the configuration value
        with the environment variable's value.
        """
        for section in self._config.sections():
            for key in self._config[section]:
                env_var = (
                    f"{self.env_prefix}__{section}_{key}".upper()
                    if self.env_prefix
                    else f"{section}_{key}".upper()
                )
                if env_var in os.environ:
                    _value = os.environ[env_var]
                    logger.debug(f"Override {section=}, {key=} with {env_var}")
                    self._config.set(section=section, option=key, value=_value)
                else:
                    logger.debug(f"Environment variable {env_var} not set")

        _default_section = self._config.default_section
        for key in self._config.defaults():
            env_var = (
                f"{self.env_prefix}_{key}".upper()
                if self.env_prefix
                else f"{key}".upper()
            )
            if env_var in os.environ:
                _value = os.environ[env_var]
                logger.debug(
                    f"Override section={_default_section}, {key=} with {env_var}"
                )
                self._config.set(section=_default_section, option=key, value=_value)
            else:
                logger.debug(f"Environment variable {env_var} not set")

    def _apply_overrides(self):
        """
        Internal method to apply direct overrides passed during initialization.
        """
        for key, value in self.overrides.items():
            if "__" in key:
                section, option = key.split("__", 1)
            else:
                section, option = self._config.default_section, key

            if section.upper() == self._config.default_section.upper():
                logger.debug(
                    f"Override default section, {option=} with direct argument"
                )
                self._config.set(section=section, option=option, value=value)
            else:
                if not self._config.has_section(section):
                    self._config.add_section(section)
                logger.debug(f"Override {section=}, {option=} with direct argument")
                self._config.set(section=section, option=option, value=value)

    def read(
        self,
        filenames: StrOrBytesPath | Iterable[StrOrBytesPath],
        encoding: str | None = None,
    ) -> configparser.ConfigParser:
        """
        Read configuration from one or more files and override with environment
        variables if set.

        This method is a wrapper around :py:meth:`configparser.ConfigParser.read` that
        reads the specified filenames in order. After reading the files, it overrides
        configuration values with any corresponding environment variables and direct
        overrides passed during initialization.

        :param filenames: A single filename or an iterable of filenames to read.
        :type filenames: :py:class:`_typeshed.StrOrBytesPath` or
            Iterable[:py:class:`_typeshed.StrOrBytesPath`]
        :param encoding: The encoding to use for reading the files, defaults to None.
        :type encoding: str, optional
        :return: The :py:class:`configparser.ConfigParser` object with the loaded and
            possibly overridden configuration.
        :rtype: :py:class:`configparser.ConfigParser`
        """
        self._config.read(filenames=filenames, encoding=encoding)
        self._find_and_override()
        self._apply_overrides()
        return self.config

    @property
    def config(self) -> configparser.ConfigParser:
        """
        Property to access the configuration.

        :return: The :py:class:`configparser.ConfigParser` object
            with the configuration.
        :rtype: :py:class:`configparser.ConfigParser`
        """
        return self._config



__version__ = "0.1.0"

import logging

from configparser_override.configparser_override import ConfigParserOverride

logging.getLogger(__name__).addHandler(logging.NullHandler())

__all__ = ["__version__", "ConfigParserOverride"]



import configparser
import os

import pytest

from configparser_override import ConfigParserOverride, __version__

TEST_ENV_PREFIX = "TEST"


@pytest.fixture(autouse=True)
def clear_env():
    # Clear environment variables before and after each test
    keys_to_clear = [
        key
        for key in os.environ
        if key.startswith(TEST_ENV_PREFIX) or key == "DEFAULT_KEY"
    ]

    for key in keys_to_clear:
        del os.environ[key]

    yield

    for key in keys_to_clear:
        if key in os.environ:
            del os.environ[key]


@pytest.fixture
def config_file(tmp_path):
    config_content = """
    [section1]
    key1 = value1
    key2 = value2

    [section2]
    key3 = value3
    """
    config_path = tmp_path / "config.ini"
    config_path.write_text(config_content)
    return str(config_path)


def test_initialization():
    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    assert parser.env_prefix == TEST_ENV_PREFIX
    assert isinstance(parser.config, configparser.ConfigParser)


def test_read_config_file(config_file):
    parser = ConfigParserOverride()
    config = parser.read(filenames=config_file)

    assert config["section1"]["key1"] == "value1"
    assert config["section1"]["key2"] == "value2"
    assert config["section2"]["key3"] == "value3"


def test_env_override(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}__SECTION1_KEY1", "override1")
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}__SECTION2_KEY3", "override3")

    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    config = parser.read(filenames=config_file)

    assert config["section1"]["key1"] == "override1"
    assert config["section1"]["key2"] == "value2"  # Not overridden
    assert config["section2"]["key3"] == "override3"


def test_default_section_override(monkeypatch, tmp_path):
    config_content = """
    [DEFAULT]
    default_key = default_value
    """
    config_path = tmp_path / "default_config.ini"
    config_path.write_text(config_content)

    monkeypatch.setenv(f"{TEST_ENV_PREFIX}_DEFAULT_KEY", "override_default")

    parser = ConfigParserOverride(env_prefix=TEST_ENV_PREFIX)
    config = parser.read(filenames=str(config_path))

    assert config.defaults()["default_key"] == "override_default"


def test_version_exits():
    assert isinstance(__version__, str)


def test_direct_override_with_env(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}__SECTION1_KEY1", "env_override_value1")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, section1__key1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["section1"]["key1"] == "direct_override_value1"
    assert config["section1"]["key2"] == "value2"  # Not overridden
    assert config["section2"]["key3"] == "value3"  # Not overridden


def test_direct_override_with_file(config_file):
    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, section1__key1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["section1"]["key1"] == "direct_override_value1"
    assert config["section1"]["key2"] == "value2"  # Not overridden
    assert config["section2"]["key3"] == "value3"  # Not overridden


def test_direct_override_with_env_and_file(monkeypatch, config_file):
    monkeypatch.setenv(f"{TEST_ENV_PREFIX}__SECTION1_KEY2", "env_override_value2")

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, section1__key1="direct_override_value1"
    )
    config = parser.read(filenames=config_file)

    assert config["section1"]["key1"] == "direct_override_value1"
    assert config["section1"]["key2"] == "env_override_value2"
    assert config["section2"]["key3"] == "value3"  # Not overridden


def test_direct_override_default_section(tmp_path):
    config_content = """
    [DEFAULT]
    default_key = default_value
    """
    config_path = tmp_path / "default_config.ini"
    config_path.write_text(config_content)

    parser = ConfigParserOverride(
        env_prefix=TEST_ENV_PREFIX, default_key="direct_override_default_value"
    )
    config = parser.read(filenames=str(config_path))

    assert config.defaults()["default_key"] == "direct_override_default_value"
